// Meta info
ENT.Type                     = "anim"
ENT.Base                     = "base_anim"

ENT.PrintName                = "Propulsion Gel"
ENT.Author                    = "Overv"
ENT.Information            = "Portal 2's propulsion gel."
ENT.Category                = "Best of Toybox"

ENT.Spawnable                = true
ENT.AdminSpawnable    = true

ENT.RenderGroup         = RENDERGROUP_TRANSLUCENT

// Initialization
function ENT:Initialize()
    if ( SERVER ) then        
        self:SetModel( "models/props_junk/MetalBucket01a.mdl" )

        self:PhysicsInit( SOLID_VPHYSICS )
        self:SetMoveType( MOVETYPE_VPHYSICS )
        self:SetSolid( SOLID_VPHYSICS )

        self:GetPhysicsObject():Wake()

        // Indicates whether this is still a can or already a spot
        self:SetDTInt( 0, 0 )
        
        self:SetTrigger( true )
        
        // Keep track of which player is holding the entity
        hook.Add( "PhysgunPickup", self:EntIndex() + CurTime(), function( ply, ent )
            if ( ent == self ) then
                if ( self:GetDTInt( 0 ) == 0 ) then
                    self.HoldingPlayer = ply
                else
                    return false
                end
            end
        end )
        
        hook.Add( "PhysgunDrop", self.EntIndex() + CurTime(), function( ply, ent )
            if ( ply == self.HoldingPlayer ) then
                self.HoldingPlayer = nil
            end
        end )
    else
        self.Shake = 0

        self:SetRenderBounds( Vector( -128, -128, -128 ), Vector( 128, 128, 128 ) )
    end
    
    self.Height = 0
end

// Draw the paint can or the paint spot
function ENT:Draw()
    // Load materials
    if ( !self.PaintMaterial ) then
        self.PaintMaterial = Material( "models/debug/debugwhite" )
        self.BlobMaterial = Material( "sprites/sent_ball" )
        self.SplashMaterial = CreateMaterial( "BluePaintMaterial", "UnlitGeneric", {
            ["$basetexture"] = "decals/decal_paintsplatterblue001",
            ["$vertexcolor"] = 1,
            ["$vertexalpha"] = 1,
            ["$model"] = 1
        } )
    end

    // Bucket phase
    if ( self:GetDTInt( 0 ) == 0 ) then
        // Calculate shake
        if ( self:GetVelocity():Length() > self.Shake * 5 * 80 ) then
            self.Shake = math.min( self:GetVelocity():Length(), 80 ) / 80 * 0.2
        end
        self.Shake = self.Shake - FrameTime() * self.Shake / 3

        // First draw regular bucket.
        self:SetModelScale( Vector( 1, 1, 1 ) )
        self:DrawModel()

        // Secondly draw the paint in the bucket.
        // The height is determined by the shaking and the angles of the bucket.
        render.SuppressEngineLighting( true )
        render.SetColorModulation( 237/255, 105/255, 49/255 )
        render.SetBlend( 1 )
        SetMaterialOverride( self.PaintMaterial )

        self:SetModelScale( Vector( 0.9, 0.9, 0.5 + math.min( self.Height * 0.67 + math.sin( CurTime() * self.Shake * 4 * math.max( 0, self:GetAngles():Up().z ) ) / 2 * self.Shake, 0.47 ) ) )
        self:DrawModel()

        render.SuppressEngineLighting( false )
        render.SetColorModulation( 1, 1, 1 )
        render.SetBlend( 1 )
        SetMaterialOverride( 0 )

        // Falling phase
    elseif ( self:GetDTInt( 0 ) == 1 ) then        
        // Draw blob
        render.SetMaterial( self.BlobMaterial )
        render.DrawSprite( self:GetPos(), 16, 32, Color( 237, 105, 49, 255 ) )

        // Splash phase
    elseif ( self:GetDTInt( 0 ) == 2 ) then
        // Draw splash

        local ang = self:GetDTAngle( 0 )

        local mat = Matrix()
        mat:Translate( self:GetPos() - 3 * ang:Forward() )
        mat:Scale( Vector( 128, 128, 128 ) )

        render.SuppressEngineLighting( true )

        render.SetMaterial( self.SplashMaterial )

        for i = 1, 5 do
            cam.PushModelMatrix( mat )
            mesh.Begin( MATERIAL_QUADS, 1 )
            mesh.Position( -1 * ang:Right() - ang:Up() )
            mesh.Normal( ang:Forward() )
            mesh.Color( 237, 105, 49, 255 )
            mesh.TexCoord( 0, 0, 0 )
            mesh.AdvanceVertex()

            mesh.Position( ang:Right() - ang:Up() )
            mesh.Normal( ang:Forward() )
            mesh.Color( 237, 105, 49, 255 )
            mesh.TexCoord( 0, 1, 0 )
            mesh.AdvanceVertex()

            mesh.Position( ang:Right() + ang:Up() )
            mesh.Normal( ang:Forward() )
            mesh.Color( 237, 105, 49, 255 )
            mesh.TexCoord( 0, 1, 1 )
            mesh.AdvanceVertex()

            mesh.Position( -1 * ang:Right() + ang:Up() )
            mesh.Normal( ang:Forward() )
            mesh.Color( 237, 105, 49, 255 )
            mesh.TexCoord( 0, 0, 1 )
            mesh.AdvanceVertex()
            mesh.End()
            cam.PopModelMatrix()
        end

        render.SuppressEngineLighting( false )
    end
end

// Predict both clientside and serverside when the paint will fall out of the bucket
function ENT:Think()
    if ( self:GetDTInt( 0 ) == 0 ) then
        self.Height = math.Clamp( self.Height - ( self:GetAngles():Up().z - 0.1 ) * 1/60 / 2, 0, 1 )

        if ( SERVER and self.Height > 0.9 or ( self:GetVelocity():GetNormal():Dot( self:GetAngles():Up() ) > 0.7 ) and self:GetVelocity():Length() > 400 ) then
            self:Drop()
        end
    elseif ( self:GetDTInt( 0 ) == 1 and CLIENT ) then
        // Make bucket fade out
        if ( !self.BucketTimeout ) then
            self.FadingBucket = ClientsideModel( "models/props_junk/MetalBucket01a.mdl" )
            self.FadingBucket:SetPos( self:GetPos() )
            self.FadingBucket:SetAngles( self:GetAngles() )

            self.BucketTimeout = CurTime() + 1
        end
    end

    // Fade bucket
    if ( self.BucketTimeout ) then
        if ( CurTime() < self.BucketTimeout ) then
            self.FadingBucket:SetColor( 255, 255, 255, ( self.BucketTimeout - CurTime() ) * 255 )
        elseif ( self.FadingBucket:IsValid() ) then
            self.FadingBucket:Remove()
        end
    end

    self:NextThink( CurTime() + 1/60 )
    return true
end

// Drop the paint
function ENT:Drop()     
    timer.Simple( 0.1, function()
        self:PhysicsInitBox( Vector( -128, -128, -4 ), Vector( 128, 128, 16 ) )
        self:SetCollisionBounds( Vector( -128, -128, -4 ), Vector( 128, 128, 16 ) )
        self:GetPhysicsObject():SetMass( 300 )

        self:SetCollisionGroup( COLLISION_GROUP_WEAPON )
        self:SetSolid( SOLID_BBOX )

        self:GetPhysicsObject():EnableMotion( true )
        self:GetPhysicsObject():Wake()
    end )

    // The old physics model is going to be invalid, so make sure no-one's holding it.
    if ( self.HoldingPlayer and self.HoldingPlayer:IsValid() ) then self.HoldingPlayer:ConCommand( "-attack" ) end

    self:SetDTInt( 0, 1 )
end

// Freeze the paint spot as soon as it hits something (usually the ground).
function ENT:PhysicsCollide( data, phys )
    if ( self:GetDTInt( 0 ) == 1 ) then
        self:EmitSound( "physics/flesh/flesh_squishy_impact_hard3.wav" )
        self:DrawShadow( false )

        self:SetDTInt( 0, 2 )
        self:SetDTAngle( 0, Vector( 0, 0, 1 ):Angle() )

        self:SetAngles( data.HitNormal:Angle() )
        self:SetMoveType( MOVETYPE_NONE )

        phys:EnableMotion( false )
    end
end

function ENT:StartTouch( ent )
    if ( ent:IsPlayer() ) then
        ent.RunSpeed = ent.RunSpeed or 500
        
        timer.Destroy( "PropulsionTimeout" )
        
        if ( ent.RunSpeed == 500 ) then
            ent:EmitSound( "player/suit_sprint.wav", 500 )
        end
        
        ent.Propulsion = true
        ent.RunSpeed = math.min( ent.RunSpeed + 500, 2000 )
        ent:SetRunSpeed( ent.RunSpeed )
        ent:SetWalkSpeed( ent.RunSpeed / 2 )
    end
end

function ENT:EndTouch( ent )
    if ( ent:IsPlayer() ) then
        ent.Propulsion = false
        
        timer.Create( "PropulsionTimeout" .. ent:EntIndex(), 0.3, 1, function()
            ent.RunSpeed = 500
            ent:SetRunSpeed( 500 )
            ent:SetWalkSpeed( 250 )
        end )
    end
end

// Alter footstep sound
hook.Add( "PlayerFootstep", "PropulsionPaintFootstep", function( ply, ent )
    if ( ply.Propulsion ) then
        ply:EmitSound( "player/footsteps/mud" .. math.random( 1, 4 ) .. ".wav" )
        return true
    end
end )

// A player starts with the default speed
hook.Add( "PlayerSpawn", "DefaultPropulsionSpeed", function( ply )
    ply.Propulsion = false
    ply.RunSpeed = 500
    ply:SetRunSpeed( 500 )
    ply:SetWalkSpeed( 250 )
end )

// Prevent spots from colliding
hook.Add( "ShouldCollide", "NoPaintCollide", function( ent1, ent2 )
    if ( ent1:GetClass() == ClassName and ent2:GetClass() == ClassName ) then
        return false
    end
end )

// Takes care of spawning
function ENT:SpawnFunction( ply, tr )
    if ( !tr.Hit ) then return end

    local pos = tr.HitPos + tr.HitNormal * 16

    local ent = ents.Create( ClassName )
    ent:SetPos( pos )
    ent:Spawn()
    ent:Activate()

    return ent
end